/* 	$Id: vmeasure.c,v 1.7 2004/06/02 23:45:33 jose Exp $ */
/*
 * Copyright (c) 2003-2004 Jose Nazario <jose@monkey.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Jose Nazario.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <sys/param.h>
#include <sys/stat.h>
#include <sys/types.h>

#include <err.h>
#include <errno.h>
#include <pthread.h>
#include <signal.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sysexits.h>
#include <time.h>
#include <unistd.h>
#include <util.h>

#include "libmilter/mfapi.h"
#include "sys/queue.h"

TAILQ_HEAD(tailhead, entry) head;
struct tailhead *headp;

struct entry {
	char            	hostname[256];	/* host's name */
	char            	mail_from[256]; /* account name */
	char			helo[256]; 	/* from HELO/EHLO command */ 
	time_t          	time_seen;	/* last time seen, seconds */ 
	u_int			interval;	/* average time interval */
	u_int			nobs;		/* number of observations */
	TAILQ_ENTRY(entry)	entries;	/* next entry */
};

static pthread_mutex_t le_mutex = PTHREAD_MUTEX_INITIALIZER;

int
average(int nobs, int curavg, int newval)
{
	int 	curtotal;

	curtotal = nobs * curavg;
	curtotal += newval;
	return ((int)(curtotal/(nobs+1)));
}

void
dumpstats(int sig) 
{
	FILE 	       *statfile;
	time_t		now;
	static struct entry   *np;

	now = time(NULL);

	if (pthread_mutex_lock(&le_mutex) != 0) {
		fprintf(stderr, "dumpstats cannot lock, bleh\n");
		return;		/* XXX */
	}

	if ((statfile = fopen("/var/tmp/vmeasure.stats", "w")) == NULL) {
		warn("cannot dump statistics.");
		return;
	}

	fprintf(statfile, "# vmeasure statistics (generated at %d)\n", now);
	fprintf(statfile, "# type\tinfo\tavg interval\tnum observations\n");

	TAILQ_FOREACH(np, &head, entries) {
		if (strlen(np->hostname) > 1) 
			fprintf(statfile, "host\t%s\t%d\t%d\n", 
				np->hostname, np->interval, np->nobs);
		if (strlen(np->mail_from) > 1)
			fprintf(statfile, "address\t%s\t%d\t%d\n", 
				np->mail_from, np->interval, np->nobs);
		if (strlen(np->helo) > 1)
			fprintf(statfile, "helo\t%s\t%d\t%d\n", 
				np->helo, np->interval, np->nobs);
	}
	fclose(statfile);
	pthread_mutex_unlock(&le_mutex);
	return;
}

sfsistat
w_connect(SMFICTX *ctx, char *hostname, _SOCK_ADDR *hostaddr)
{
	time_t			now;
	static struct entry    *np, *n2;
	
	now = time(NULL);

	if (pthread_mutex_lock(&le_mutex) != 0) {
		fprintf(stderr, "w_connect cannot lock, bleh\n");
		return SMFIS_REJECT;    /* XXX */
	}

	/* has this entry been seen before? */
	TAILQ_FOREACH(np, &head, entries) {
		if ((np->hostname != NULL) &&
		    (strncasecmp(hostname, np->hostname, 
			     strlen(hostname)) == 0)) {
			np->interval = average(np->nobs, np->interval,
					      now - np->time_seen);
			np->nobs += 1;
			np->time_seen = now;
			pthread_mutex_unlock(&le_mutex);
			return SMFIS_CONTINUE;
		}
	}

	/* new entry */
	n2 = malloc(sizeof(struct entry));
	n2->time_seen = now;

	snprintf(n2->hostname, 256, "%s", hostname);
	snprintf(n2->helo, 256, "%s", "");
	snprintf(n2->mail_from, 256, "%s", "");
	n2->time_seen = now;
	n2->nobs = 1;
	n2->interval = 0;
	TAILQ_INSERT_TAIL(&head, n2, entries);

	pthread_mutex_unlock(&le_mutex);
	return SMFIS_CONTINUE;
}

sfsistat
w_helo(SMFICTX * ctx, char *helohost)
{
	time_t          	now;
	static struct entry    *np, *n2;

	now = time(NULL);

	if (pthread_mutex_lock(&le_mutex) != 0) {
		fprintf(stderr, "w_connect cannot lock, bleh\n");
		return SMFIS_REJECT;    /* XXX */
	}

	/* has this entry been seen before? */
	TAILQ_FOREACH(np, &head, entries) {
		if ((np->helo != NULL) &&
		    (strncasecmp(helohost, np->helo, 
			     strlen(helohost)) == 0)) {
			np->interval = average(np->nobs, np->interval,
					      now - np->time_seen);
			np->nobs += 1;
			np->time_seen = now;
			pthread_mutex_unlock(&le_mutex);
			return SMFIS_CONTINUE;
		}
	}

	/* new entry */
	n2 = malloc(sizeof(struct entry));
	n2->time_seen = now;

	snprintf(n2->helo, 256, "%s", helohost);
	snprintf(n2->hostname, 256, "%s", "");
	snprintf(n2->mail_from, 256, "%s", "");
	n2->time_seen = now;
	n2->nobs = 1;
	n2->interval = 0;
	TAILQ_INSERT_TAIL(&head, n2, entries);

	pthread_mutex_unlock(&le_mutex);
	return SMFIS_CONTINUE;
}

sfsistat
w_envfrom(SMFICTX * ctx, char **argv)
{
	time_t          	now;
	char                   *mailaddr = smfi_getsymval(ctx, "{mail_addr}");
	static struct entry    *np, *n2;

	now = time(NULL);

	if (pthread_mutex_lock(&le_mutex) != 0) {
		fprintf(stderr, "w_connect cannot lock, bleh\n");
		return SMFIS_REJECT;    /* XXX */
	}

	/* walk the list first ... */
	TAILQ_FOREACH(np, &head, entries) {
		if (strncasecmp(mailaddr, np->mail_from,
			    sizeof(np->mail_from)) == 0) {
			np->interval = average(np->nobs, np->interval,
					      now - np->time_seen);
			np->nobs += 1;
			np->time_seen = now;
			pthread_mutex_unlock(&le_mutex);
			return SMFIS_CONTINUE;
		}
	}

	/* new entry */
	n2 = malloc(sizeof(struct entry));
	n2->time_seen = now;

	snprintf(n2->mail_from, 256, "%s", mailaddr);
	snprintf(n2->helo, 256, "%s", "");
	snprintf(n2->hostname, 256, "%s", "");
	n2->time_seen = now;
	n2->nobs = 1;
	n2->interval = 0;
	TAILQ_INSERT_TAIL(&head, n2, entries);

	pthread_mutex_unlock(&le_mutex);
	return SMFIS_CONTINUE;
}

struct smfiDesc smfilter =
{
	"vmeasure",
	SMFI_VERSION,
	SMFIF_ADDHDRS,
	w_connect,
	w_helo,
	w_envfrom,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
};

static void
usage(char *prog)
{
	fprintf(stderr, 
	    "usage: %s -p sockaddr\n", prog);
	exit(1);
}

/*
 * a complementary tool to vthrottle which is used to measure the
 * average interval in seconds between SMTP observations. useful to
 * build an accurate whitelist for vthrottle or to gather statistics
 * about your SMTP server.
 */

int
main(int argc, char **argv)
{
	int             c, setconn = 0, retval;
	extern char    *optarg;
	struct entry   *n1;

	TAILQ_INIT(&head); 

	while ((c = getopt(argc, argv, "i:p:w:h")) != -1) {
		switch (c) {
		case 'p':
			/* XXX - need to check for this file first */
			if (smfi_setconn(optarg) == MI_FAILURE)
				err(1, "socket connect failure.");
			if (strncmp(optarg, "unix:", 5) == 0)
				unlink(optarg + 5);
			if (strncmp(optarg, "local:", 6) == 0)
				unlink(optarg + 6);
			setconn = 1;
			break;
		case 'h':
		default:
			usage(argv[0]);
			break;
		}
	}

	if (!setconn) 
		usage(argv[0]);

	signal(SIGUSR1, dumpstats);

	n1 = malloc(sizeof(struct entry));
	n1->time_seen = 100;
	n1->interval = 0;
	snprintf(n1->mail_from, 256, "nobody");
	TAILQ_INSERT_HEAD(&head, n1, entries);

	if (smfi_register(smfilter) == MI_FAILURE) 
		err(1, "initialization failure.");

	retval = smfi_main();
	dumpstats(0);
	pthread_mutex_destroy(&le_mutex);
	return retval;
}
