/* $Id: vthrottle.c,v 1.29 2004/06/02 23:51:18 jose Exp $ */
/*
 * Copyright (c) 2003-2004 Jose Nazario <jose@monkey.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Jose Nazario.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <sys/param.h>
#include <sys/stat.h>
#include <sys/types.h>

#include <err.h>
#include <errno.h>
#include <pthread.h>
#include <signal.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sysexits.h>
#include <syslog.h>
#include <time.h>
#include <unistd.h>
#include <util.h>

#include "libmilter/mfapi.h"
#include "sys/queue.h"

TAILQ_HEAD(tailhead, entry) head;
struct tailhead *headp;

struct entry {
	char            	hostname[256];	/* host's name */
	char            	mail_from[256]; /* account name */
	char			helo[256]; 	/* from HELO/EHLO command */ 
	time_t          	time_seen;	/* last time seen, seconds */ 
	u_int			interval;	/* expected time interval */
	u_int			throttled;	/* times blocked */
	u_int			flags;		/* info flags */
#define V_WHITELIST		0x001
	TAILQ_ENTRY(entry)	entries;	/* next entry */
};

int			interval = 60, expire = 0;
static pthread_mutex_t  le_mutex = PTHREAD_MUTEX_INITIALIZER;

void
dumpstats(int sig) 
{
	FILE 	      	       *statfile;
	time_t			now;
	static struct entry    *np;

	now = time(NULL);

	if (pthread_mutex_lock(&le_mutex) != 0) {
		warn("dumpstats cannot lock, bleh\n");
		return;		/* XXX */
	}

	if ((statfile = fopen("/var/tmp/vthrottle.stats", "a")) == NULL) {
		warn("statfile error in dumpstats.");
		return;
	}

	fprintf(statfile, "\nvthrottle statistics (%d),\n", now);

	/* print info for each entry */
	TAILQ_FOREACH(np, &head, entries) {
		if (strlen(np->hostname) > 1) {
			fprintf(statfile, 
			    "SMTP host: %s, expected interval: %d, last seen: %d, throttled %d times\n",
			    np->hostname, np->interval, np->time_seen, np->throttled);
		} else if (strlen(np->mail_from) > 1) {
			fprintf(statfile, 
			    "address: %s, expected interval: %d, last seen: %d, throttled %d times\n",
			    np->mail_from, np->interval, np->time_seen, np->throttled);
		} else if (strlen(np->helo) > 1) {
			fprintf(statfile, 
			    "HELO host: %s, expected interval: %d, last seen: %d, throttled %d times\n",
			    np->helo, np->interval, np->time_seen, np->throttled);
		}
	}

	fclose(statfile);
	pthread_mutex_unlock(&le_mutex);
	return;
}

/* 
   builds a whitelist of hosts and addresses that can send unlimited.
   returns 0 on success, -1 on failure.
*/

static int
parse_whitelist(char *filename)
{
	FILE	              *file;
	static struct entry   *n2;
	char 		      *input;

	if ((file = fopen(filename, "r")) == NULL) {
		warn("Failed to open whitelist file");
		return(-1);
	}

	for (; (input = fparseln(file, NULL, NULL, NULL, 0)); free(input)) {
		int	i = 0;
		char   *next, *entry[3];

		if (input == NULL)
			break;

		next = input;
		for (; i < 3 && (entry[i] = strsep(&next, " \t")) != NULL;)
			if (*entry[i] != '\0')
				i++;

		/* do we have a type/value/interval triplet? */
		if (i >= 2) {
			n2 = malloc(sizeof(struct entry));
			n2->time_seen = 0;
			n2->flags = V_WHITELIST;
			if (strncasecmp(entry[0], "helo", 4) == 0) {
				snprintf(n2->hostname, 256, "%s", "");
				snprintf(n2->mail_from, 256, "%s", "");
				snprintf(n2->helo, 256, "%s", entry[1]);
			}
			if (strncasecmp(entry[0], "host", 4) == 0) {
				snprintf(n2->hostname, 256, "%s", entry[1]);
				snprintf(n2->mail_from, 256, "%s", "");
				snprintf(n2->helo, 256, "%s", "");
			}
			if (strncasecmp(entry[0], "address", 7) == 0)  {
				snprintf(n2->hostname, 256, "%s", "");
				snprintf(n2->mail_from, 256, "%s", entry[1]);
				snprintf(n2->helo, 256, "%s", "");
			}
			n2->interval = atoi(entry[2]);
			TAILQ_INSERT_TAIL(&head, n2, entries);
		}
	}

	fclose(file);
	return(0);
}

void
prune(time_t now, struct entry *np)
{
	char		       *description;
	struct syslog_data 	sdata = SYSLOG_DATA_INIT;

	if (!expire)
		return;

	if (!np)
		return;

	/* not too old */
	if ((now - np->time_seen) <= expire)
		return;

	/* don't prnue whitelist entries */
	if (np->flags & V_WHITELIST)
		return;

	if (strlen(np->hostname) > 1)
		asprintf(&description, "connecting host %s", np->hostname);	
	else if (strlen(np->helo) > 1)
		asprintf(&description, "HELO host %s", np->helo);
	else if (strlen(np->mail_from) > 1)
		asprintf(&description, "from %s", np->mail_from);
	else
		asprintf(&description, "unknown");

	syslog_r(LOG_WARNING, &sdata,
		"removing expired entry for %s, age %d is older than %d\n",
		description, now - np->time_seen, expire);

	free(description);

	/* remove stale entries */
	TAILQ_REMOVE(&head, np, entries);
	free(np);
}


sfsistat
w_connect(SMFICTX *ctx, char *hostname, _SOCK_ADDR *hostaddr)
{
	static time_t		now;
	static struct entry    *np, *n2;
	
	now = time(NULL);

	if (smfi_setreply(ctx, "451", "4.7.1", "Please try again later") == MI_FAILURE) { 
		struct syslog_data sdata = SYSLOG_DATA_INIT;
		syslog_r(LOG_WARNING, &sdata, "w_connect, cannot set reply to 451");
	}

	if (pthread_mutex_lock(&le_mutex) != 0) {
		warn("w_connect cannot lock, bleh\n");
		return SMFIS_TEMPFAIL;    /* XXX */
	}

	/* has this entry been seen before? */
	TAILQ_FOREACH(np, &head, entries) {
		if ((np->hostname != NULL) &&
		    (strncasecmp(hostname, np->hostname, 
		     strlen(hostname)) == 0)) {
			/* no more than one connection per interval ... */
			if ((now - np->time_seen) < (np->interval)) {
				struct syslog_data sdata = SYSLOG_DATA_INIT;

				syslog_r(LOG_WARNING, &sdata,
				    "throttling connecting host %s: observed interval: %d, expected: %d\n", hostname, now - np->time_seen, np->interval);
				np->throttled += 1;
				pthread_mutex_unlock(&le_mutex);
				return SMFIS_TEMPFAIL;
			} else {
				np->time_seen = now;
				pthread_mutex_unlock(&le_mutex);
				return SMFIS_CONTINUE;
			}
		} else
			prune(now, np);
	}

	/* new entry */
	n2 = malloc(sizeof(struct entry));
	if (n2 == NULL) 
		return SMFIS_CONTINUE; 		/* XXX - don't fail closed */
	n2->time_seen = now;
	n2->flags = 0;

	snprintf(n2->hostname, 256, "%s", hostname);
	snprintf(n2->helo, 256, "%s", "");
	snprintf(n2->mail_from, 256, "%s", "");
	n2->interval = interval;
	TAILQ_INSERT_TAIL(&head, n2, entries);

	pthread_mutex_unlock(&le_mutex);
	return SMFIS_CONTINUE;
}

sfsistat
w_helo(SMFICTX * ctx, char *helohost)
{
	static time_t          	now;
	static struct entry    *np, *n2;

	now = time(NULL);

	if (smfi_setreply(ctx, "451", "4.7.1", "Please try again later") == MI_FAILURE) { 
		struct syslog_data sdata = SYSLOG_DATA_INIT;
		syslog_r(LOG_WARNING, &sdata, "w_helo, cannot set reply to 451");
	}

	if (pthread_mutex_lock(&le_mutex) != 0) {
		warn("w_helo cannot lock, bleh\n");
		return SMFIS_TEMPFAIL;    /* XXX */
	}

	/* has this entry been seen before? */
	TAILQ_FOREACH(np, &head, entries) {
		if ((np->helo != NULL) &&
		    (strncasecmp(helohost, np->helo, 
		     strlen(helohost)) == 0)) {

			/* no more than one connection per interval ... */
			if ((now - np->time_seen) < (np->interval)) {
				struct syslog_data sdata = SYSLOG_DATA_INIT;

				syslog_r(LOG_WARNING, &sdata,
				    "throttling host %s HELO host %s: observed interval: %d, expected: %d\n", np->hostname, helohost, now - np->time_seen, np->interval);
				np->throttled += 1;
				pthread_mutex_unlock(&le_mutex);
				return SMFIS_TEMPFAIL;
			} else {
				np->time_seen = now;
				pthread_mutex_unlock(&le_mutex);
				return SMFIS_CONTINUE;
			}
		} else
			prune(now, np);
	}

	/* new entry */
	n2 = malloc(sizeof(struct entry));
	if (n2 == NULL)
		return SMFIS_CONTINUE; 	/* XXX - but don't fail closed */
	n2->time_seen = now;
	n2->flags = 0;

	snprintf(n2->helo, 256, "%s", helohost);
	snprintf(n2->hostname, 256, "%s", "");
	snprintf(n2->mail_from, 256, "%s", "");
	n2->interval = interval;
	TAILQ_INSERT_TAIL(&head, n2, entries);

	pthread_mutex_unlock(&le_mutex);
	return SMFIS_CONTINUE;
}

sfsistat
w_envfrom(SMFICTX * ctx, char **argv)
{
	static time_t          	now;
	static char            *mailaddr;
	static struct entry    *np, *n2;

	now = time(NULL);

	if (pthread_mutex_lock(&le_mutex) != 0) {
		warn("w_from cannot lock, bleh\n");
		return SMFIS_TEMPFAIL;    /* XXX */
	}

	mailaddr = smfi_getsymval(ctx, "{mail_addr}");

	/* walk the list first ... */
	TAILQ_FOREACH(np, &head, entries) {
		if (strncasecmp(mailaddr, np->mail_from,
		    sizeof(np->mail_from)) == 0) {

			/* no more than one connection per interval ... */
			if ((now - np->time_seen) < (np->interval)) {
				struct syslog_data sdata = SYSLOG_DATA_INIT;

				syslog_r(LOG_WARNING, &sdata,
				    "throttling mail address %s: observed interval: %d, expected: %d\n", mailaddr, now - np->time_seen, np->interval);
				np->throttled += 1;
				pthread_mutex_unlock(&le_mutex);
				return SMFIS_TEMPFAIL;
			} else {
				np->time_seen = now;
				pthread_mutex_unlock(&le_mutex);
				return SMFIS_CONTINUE;
			}
		} else
			prune(now, np);
	}

	/* new entry */
	n2 = malloc(sizeof(struct entry));
	if (n2 == NULL)
		return SMFIS_CONTINUE;	/* XXX - but don't fail closed */
	n2->time_seen = now;
	n2->flags = 0;

	snprintf(n2->mail_from, 256, "%s", mailaddr);
	snprintf(n2->helo, 256, "%s", "");
	snprintf(n2->hostname, 256, "%s", "");
	n2->interval = interval;
	TAILQ_INSERT_TAIL(&head, n2, entries);

	pthread_mutex_unlock(&le_mutex);
	return SMFIS_CONTINUE;
}

struct smfiDesc smfilter =
{
	"vthrottle",
	SMFI_VERSION,
	SMFIF_ADDHDRS,
	w_connect,
	w_helo,
	w_envfrom,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
};

static void
usage(char *prog)
{
	fprintf(stderr, 
	    "usage: %s -p sockaddr [-i interval] [-w whitelist] [-x expire]\n", prog);
	exit(1);
}

/*
 * an implementation of m. williamson's rate throttling mechanism for
 * mail bourne worms and viruses. basically we keep track of what hosts
 * and what senders have been sending mail. if they attempt to send mail
 * from one machine or using one address more than once in any interval
 * we back them off by sending a (temporary) failure. well behaved hosts
 * are expected to not trip this detection/throttle ...
 *
 * references:
 * Williamson, M. HPL-2002-172
 *      HP Labs Tech Report HPL-2002-172
 * 	http://www.hpl.hp.com/techreports/2002/HPL-2002-172.html
 * Twycross, J., and Williamson, M.
 * 	in Proc. 12th USENIX Security Symposium, pp 285-294
 *	http://www.usenix.org/events/sec03/tech/twycross.html
 */

int
main(int argc, char **argv)
{
	int            		c, setconn = 0, retval;
	extern char    	       *optarg;
	struct syslog_data 	sdata = SYSLOG_DATA_INIT;
	struct entry   	       *n1;

	TAILQ_INIT(&head); 

	while ((c = getopt(argc, argv, "i:p:w:x:h")) != -1) {
		switch (c) {
		case 'i':
			interval = atoi(optarg);
			break;
		case 'p':
			/* XXX - need to check for this file first */
			if (smfi_setconn(optarg) == MI_FAILURE)
				err(1, "socket connect failure.");
			if (strncmp(optarg, "unix:", 5) == 0)
				unlink(optarg + 5);
			if (strncmp(optarg, "local:", 6) == 0)
				unlink(optarg + 6);
			setconn = 1;
			break;
		case 'w':
			if (parse_whitelist(optarg) < 0)
				err(1, "whitelist file malformed.");
			break;
		case 'x':
			expire = atoi(optarg);
			if (expire < 1)
				err(1, "expiration interval must be greater than 0.");
			if (expire < interval)
				err(1, "expiration interval must be greater than the default throttle interval.");
			break;
		case 'h':
		default:
			usage(argv[0]);
			break;
		}
	}

	if (!setconn) 
		usage(argv[0]);

	/* XXX threadsafe? */
	signal(SIGUSR1, dumpstats);

	n1 = malloc(sizeof(struct entry));
	if (n1 == NULL)
		err(1, "initialization failure in malloc.");
	n1->time_seen = 100;
	n1->interval = interval;
	snprintf(n1->mail_from, 256, "nobody");
	TAILQ_INSERT_HEAD(&head, n1, entries);

	openlog_r("vthrottle", LOG_PID | LOG_NDELAY, LOG_MAIL, &sdata);

	if (smfi_register(smfilter) == MI_FAILURE) 
		err(1, "initialization failure.");

	retval = smfi_main();
	closelog();
	pthread_mutex_destroy(&le_mutex);
	return retval;
}
